// rymshot.cpp
// 2002-03-29 23:48 
// http://litestep.waglo.com/modules/rymshot/
/*
2002-03-31, v0.17: initial release
2002-03-31, v0.18: finished docs, little cleanups
2002-04-01, v0.19: ftp upload, disable unneeded bangs, enabled bmp support
2002-04-01, v0.21: a few bugfixes (resolution, ftp folder, etc.)
2002-04-02, v0.25: new bang: rymshotactive and associated setup options
2002-04-02, v0.26: now compiles with g++
2002-04-03, v0.27: text overlay options
2002-04-03, v0.32: fixed transparency issues on nt/w2k/xp (thanks, Vulpes!)
                    "implemented" file locks
                    (Vulpes saw a problem when running ACDSee 4.0)
2002-04-03, v0.33: new options: rymshotftpnocaption and rymshotftpnothumbnail
2002-04-03, v0.36: new image file format: png
2002-04-07, v0.39: fixed included headers to make building more cross-platform
                    removed couple extra functions
                    fixed recycling problems with old configs corrupting new
                    using strnxxx instead of strxxx
                    new ftp bang to set password, cant have empty password
2002-04-10, v0.49: new option: rymshotftpuselocal use local settings for ftp
                   bugfix: jpeg quality over ftp wasn't handled (thanks, Nick
                   bugfix: jpeg thumbnail didn't have a file extension (Young)
				   new bang: !rymshotabout displays info about current setup

                               

*/

#include "../ls-b24/lsapi/lsapi.h"

#include <stdlib.h>
#include <fcntl.h>

#ifdef __CYGWIN__
#include <sys/unistd.h>
#else
#include <io.h>
#endif

#include <stdio.h>
#include <string.h>
#include <time.h>
#include <math.h>

#include <jpeglib.h>
#include <png.h>

#include <wininet.h>


#include "pngdib.h"

#ifndef CAPTUREBLT 
#define CAPTUREBLT 0x40000000
#endif

typedef enum
{
      JPEG_FORMAT
    , BMP_FORMAT
    , PNG_FORMAT
    , PNM_FORMAT
} image_format;


int screenshot_desktop(const char * filename
    , int image_width
    , int image_height
    , int quality
    , image_format format
    , bool no_caption = false
    , bool is_thumbnail = false);

///////////////////////////////////////////////////////////////


// pngdib.c
//
// PNGDIB - a mini imaging library for Win32
// By Jason Summers
// This software may be used without restriction.
//
// read_png_to_dib()
//     Read a PNG image file and convert it to a Windows "Device Independent
//     Bitmap" (DIB)
//
// write_dib_to_png()
//     Write a DIB to a PNG image file
// 
// pngdib_get_version_string(void)
//
// pngdib_get_version(void)
//

#define PNGDIB_SRC_VERSION           20101
#define PNGDIB_SRC_VERSION_STRING   "2.1.1"

#if PNGDIB_SRC_VERSION != PNGDIB_HEADER_VERSION
#error Wrong PNGDIB header file version
#endif


// This is basically a Windows-only utility with a simple-as-possible
// interface, so I'm not too concerned about allowing a 
// user-configurable screen gamma.
static const double screen_gamma = 2.2;

#ifndef uchar
#define uchar unsigned char
#endif

#define MAX_ERRMSGLEN 100

struct errstruct {
    jmp_buf *jbufp;
    char *errmsg;
};

static uchar* uncompress_dib(LPBITMAPINFO lpbmi1, int infosize, void *lpbits1)
{
    LPBITMAPINFOHEADER lpdib2;
    uchar *lpbits2;
    void *whatever;
    int linesize, bitssize;
    HBITMAP hb;
    HDC hdc;
    HGDIOBJ rvgdi;
    int rvi;
    int width,height;
    LPBITMAPINFOHEADER lpdib1;

    lpdib1=(LPBITMAPINFOHEADER)lpbmi1;
    width=lpdib1->biWidth;
    height=lpdib1->biHeight;

    linesize= (((width * lpdib1->biBitCount)+31)/32)*4;
    bitssize= linesize*height;

    lpdib2= (LPBITMAPINFOHEADER)GlobalAlloc(GPTR,infosize);
    if(!lpdib2) return NULL;

    //  create a header for the new uncompressed DIB
    CopyMemory((void*)lpdib2,(void*)lpdib1,infosize);
    lpdib2->biCompression=BI_RGB;
    lpdib2->biSizeImage=0;

    lpbits2= (uchar*)GlobalAlloc(GPTR,bitssize);
    if(!lpbits2) { GlobalFree(lpdib2); return NULL; }


    // Windows bitmap handling functions are not exactly convenient,
    // especially when trying to deal with DIBs. Every function wants
    // to convert them into DDBs. We have to play stupid games and
    // convert back and forth. This probably uses too much memory,
    // and I'm not 100% sure it is exactly correct, but it seems to
    // work for me.

    hb=CreateDIBSection(NULL,(LPBITMAPINFO)lpdib2,DIB_RGB_COLORS,&whatever,NULL,0);

    hdc=CreateCompatibleDC(NULL);
    rvgdi=SelectObject(hdc,hb);
    //SetStretchBltMode(hdc,COLORONCOLOR);
    rvi=StretchDIBits(hdc,
        0,0,width,height,
        0,0,width,height,
        lpbits1, (LPBITMAPINFO)lpdib1,
        DIB_RGB_COLORS,SRCCOPY);
    rvi=GetDIBits(hdc,hb,0,height, (LPVOID)lpbits2,
        (LPBITMAPINFO)lpdib2,DIB_RGB_COLORS);

    DeleteDC(hdc);
    DeleteObject(hb);
    GlobalFree(lpdib2);

    return lpbits2;
}


static void my_error_fn(png_structp png_ptr, const char *err_msg)
{
    struct errstruct *errinfop;
    jmp_buf *j;

    errinfop = (struct errstruct *)png_get_error_ptr(png_ptr);
    j = errinfop->jbufp;
    snprintf(errinfop->errmsg,MAX_ERRMSGLEN,"[libpng] %s",err_msg);
    //snprintf(errinfop->errmsg,MAX_ERRMSGLEN,"%s",err_msg);
    errinfop->errmsg[MAX_ERRMSGLEN-1]='\0';
    longjmp(*j, -1);
}


static void my_warning_fn(png_structp png_ptr, const char *warn_msg)
{
    return;
}



//--------------------------------------------------------------------

int write_dib_to_png(PNGD_D2PINFO *d2pp)
{
    png_structp png_ptr;
    png_infop info_ptr;
    jmp_buf jbuf;
    struct errstruct errinfo;
    png_text text_ptr[1];
    uchar *bits;
    uchar *newimage;
    RGBQUAD* dib_palette;
    int headersize, dib_bpp;
    png_uint_32 res_x, res_y;
    png_color_8 pngc8;
    int height,width;
    int palette_entries;
    int topdown;
    int dib_bytesperrow;
    int compression;
    FILE *fp;
    png_color png_palette[256];
    uchar **row_pointers;
    int i,x,y,size;
    DWORD *bitfields;
    unsigned int v;
    int bf_format;   // bitfields format identifier
    int iscompressed;
    int bfsize; // bytes in "bitfields" section
    int palsize;  // bytes in palette

    int palentsize;  // bytes in a palette entry: 3 or 4;
    BITMAPCOREHEADER *lpolddib;
    int rv;  // return code
    char *errmsg;
    char dummy_errmsg[MAX_ERRMSGLEN];

    if(d2pp->structsize != sizeof(PNGD_D2PINFO)) return PNGD_E_VERSION;
                     
    rv=PNGD_E_ERROR;  // this should always get changed before returning
    png_ptr=NULL;
    info_ptr=NULL;
    fp=NULL;
    row_pointers=NULL;
    newimage=NULL;

    if(d2pp->errmsg)
        errmsg=d2pp->errmsg;
    else
        errmsg=dummy_errmsg;
    errmsg[0]='\0';

    headersize= d2pp->lpdib->biSize;

    if(headersize<40 && headersize!=12) {
        sprintf(errmsg,"Unexpected BMP header size (%d)",headersize);
        rv=PNGD_E_BADBMP; goto abort;
    }

    if(headersize==12) {
        // This is to support an old kind of DIBs (OS/2) that aren't really
        // used anymore.
        palentsize= 3;
        lpolddib= (BITMAPCOREHEADER*) d2pp->lpdib;
        width= lpolddib->bcWidth;
        height= lpolddib->bcHeight;
        dib_bpp= lpolddib->bcBitCount;
        compression = BI_RGB;
        res_x = res_y = 0;

        // This will get adjusted later if there is a palette.
        // Not sure it's right, though. Do old DIBs always have a
        // full-sized palette?
        palette_entries=0;
    }
    else {
        palentsize=4;
        width= d2pp->lpdib->biWidth;
        height= d2pp->lpdib->biHeight;
        dib_bpp= d2pp->lpdib->biBitCount;
        compression= d2pp->lpdib->biCompression;
        res_x = d2pp->lpdib->biXPelsPerMeter;
        res_y = d2pp->lpdib->biYPelsPerMeter;
        palette_entries = d2pp->lpdib->biClrUsed;
    }

    // supposedly, if the height is negative, the top scanline is stored first
    topdown=0;
    if(height<0) {
        height= -height;
        topdown=1;
    }

    // sanity check
    if(height<1 || height>1000000 || width<1 || width>1000000) {
        sprintf(errmsg,"Unreasonable image dimensions (%dx%d)",width,height);
        rv=PNGD_E_BADBMP; goto abort;
    }

    // only certain combinations of compression and bpp are allowed
    switch(compression) {
    case BI_RGB: 
        if(dib_bpp!=1 && dib_bpp!=4 && dib_bpp!=8 && dib_bpp!=16
            && dib_bpp!=24 && dib_bpp!=32)
        {
            sprintf(errmsg,"Unsupported bit depth (%d)",dib_bpp);
            rv=PNGD_E_UNSUPP; goto abort;
        }
        break;
    case BI_RLE4:
        if(dib_bpp!=4) {
            rv=PNGD_E_UNSUPP; goto abort;
        }
        break;
    case BI_RLE8:
        if(dib_bpp!=8) {
            rv=PNGD_E_UNSUPP; goto abort;
        }
        break;
    case BI_BITFIELDS:
        if(dib_bpp!=16 && dib_bpp!=32) {
            rv=PNGD_E_UNSUPP; goto abort;
        }
        break;
    default:
        sprintf(errmsg,"Unsupported compression scheme");
        return PNGD_E_UNSUPP;
    }

    iscompressed= (compression==BI_RLE4 || compression==BI_RLE8);

    // uncompressed dibs are padded to 4-byte bondaries
    dib_bytesperrow= (((width * dib_bpp)+31)/32)*4;

    if(dib_bpp<16) {
        if(palette_entries==0) palette_entries = 1<<dib_bpp;
    }
//  if(dib_bpp>8) {
//      palette_entries=0;
//  }
//  else {
//      // 0 means it has a full palette
//      if(palette_entries==0) palette_entries = 1<<dib_bpp;
//  }

    bfsize = (compression==BI_BITFIELDS)?12:0;

    //if(compression==BI_BITFIELDS) palsize=12;
    //else palsize=palentsize*palette_entries;
    palsize=palentsize*palette_entries;

    // bounds check
    size= headersize + bfsize + palsize;
    if(d2pp->dibsize) {
        if(size>d2pp->dibsize) {
            rv=PNGD_E_BADBMP; goto abort;
        }
    }

    if(d2pp->lpbits) {
        if(d2pp->bitssize && !iscompressed) {   // bounds check
            size=dib_bytesperrow*height;
            if(size>d2pp->bitssize) { rv=PNGD_E_BADBMP; goto abort; }
        }

        bits=(uchar*)d2pp->lpbits;
    }
    else {
        // If not provided by user, assume the bits immediately
        // follow the palette.

        if(d2pp->dibsize && !iscompressed) {  // bounds check
            size= headersize+bfsize+palsize+dib_bytesperrow*height;
            if(size>d2pp->dibsize) { rv=PNGD_E_BADBMP; goto abort; }
        }

        bits= &((uchar*)(d2pp->lpdib))[headersize+bfsize+palsize];
    }

    bitfields   =  (DWORD*)  ( &((uchar*)(d2pp->lpdib))[headersize] );
    dib_palette = (RGBQUAD*) ( &((uchar*)(d2pp->lpdib))[headersize+bfsize] );

    //bitfields=(DWORD*)dib_palette;

    bf_format=0;
    if(compression==BI_BITFIELDS) {
        if(dib_bpp==16) {
            if     (bitfields[0]==0x00007c00 && bitfields[1]==0x000003e0 && 
                    bitfields[2]==0x0000001f) bf_format=11;  // 555
            else if(bitfields[0]==0x0000f800 && bitfields[1]==0x000007e0 &&
                    bitfields[2]==0x0000001f) bf_format=12;  // 565
            else { rv=PNGD_E_UNSUPP; goto abort; }
        }
        if(dib_bpp==32) {
            if     (bitfields[0]==0x00ff0000 && bitfields[1]==0x0000ff00 &&
                    bitfields[2]==0x000000ff) bf_format=21;
            else { rv=PNGD_E_UNSUPP; goto abort; }
        }
    }

    if(bf_format==0 && dib_bpp==16) bf_format=10;
    if(bf_format==0 && dib_bpp==32) bf_format=20;


    // Done analyzing the DIB, now time to convert it to PNG


    // jbuf: see comments in read_png_to_dib()
    errinfo.jbufp = &jbuf;
    errinfo.errmsg = errmsg;
    png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, (void*)(&errinfo),
        my_error_fn, my_warning_fn);
    if (!png_ptr) { rv=PNGD_E_NOMEM; goto abort; }


    info_ptr = png_create_info_struct(png_ptr);
    if (!info_ptr) {
        //png_destroy_write_struct(&png_ptr,(png_infopp)NULL);
        rv=PNGD_E_NOMEM; goto abort;
    }

    fp=fopen(d2pp->pngfn,"wb");
    if(!fp) {
        rv=PNGD_E_WRITE;
        goto abort;
    }

    png_init_io(png_ptr, fp);

    png_set_IHDR(png_ptr, info_ptr, width, height, (dib_bpp>8)?8:dib_bpp,
        (dib_bpp>8)?PNG_COLOR_TYPE_RGB:PNG_COLOR_TYPE_PALETTE,
        (d2pp->flags&PNGD_INTERLACED)?PNG_INTERLACE_ADAM7:PNG_INTERLACE_NONE,
        PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);

    // write sRGB and gAMA chunks
    if(!(d2pp->flags&PNGD_NO_GAMMA_LABEL)) {
        png_set_sRGB(png_ptr, info_ptr, PNG_sRGB_INTENT_RELATIVE);
        png_set_gAMA(png_ptr, info_ptr, 0.45455);
    }

    // For 16-bit DIBs, we get to write an sBIT chunk.
    if(bf_format==10 || bf_format==11) {
        pngc8.red= 5;  pngc8.green= 5;  pngc8.blue= 5;
        png_set_sBIT(png_ptr, info_ptr, &pngc8);
    }
    if(bf_format==12) {
        pngc8.red= 5;  pngc8.green= 6;  pngc8.blue= 5;
        png_set_sBIT(png_ptr, info_ptr, &pngc8);
    }

    // pHYs
    if(res_x>0 && res_y>0)
        png_set_pHYs(png_ptr, info_ptr, res_x, res_y, 1);


    if(palette_entries>0) {
        for(i=0;i<palette_entries;i++) {
            if(palentsize==3) {
                png_palette[i].red   = ((RGBTRIPLE*)dib_palette)[i].rgbtRed;
                png_palette[i].green = ((RGBTRIPLE*)dib_palette)[i].rgbtGreen;
                png_palette[i].blue  = ((RGBTRIPLE*)dib_palette)[i].rgbtBlue;
            }
            else {
                png_palette[i].red   = dib_palette[i].rgbRed;
                png_palette[i].green = dib_palette[i].rgbGreen;
                png_palette[i].blue  = dib_palette[i].rgbBlue;
            }
        }
        png_set_PLTE(png_ptr, info_ptr, png_palette, palette_entries);
    }

    if(dib_bpp>8)
        png_set_bgr(png_ptr);

    png_write_info(png_ptr, info_ptr);

    row_pointers=(uchar**)GlobalAlloc(GMEM_FIXED,height*sizeof(uchar*));
    if(!row_pointers) { rv=PNGD_E_NOMEM; goto abort; }


    if(dib_bpp==16 || dib_bpp==32) {

        // Special handling for these bit depths.
        // This uses a lot of memory, and could be improved by processing
        // one line at a time (but that makes it tricky to write interlaced
        // images).

        newimage=(uchar*)GlobalAlloc(GMEM_FIXED,height*width*3);
        if(!newimage) { rv=PNGD_E_NOMEM; goto abort; }

        for(y=0;y<height;y++) {
            for(x=0;x<width;x++) {
                switch(bf_format) {
                case 10:  case 11:  // 16-bit, format 555 (xRRRRRGG GGGBBBBB)
                    v= bits[y*dib_bytesperrow+x*2+0] | (bits[y*dib_bytesperrow+x*2+1]<<8);
                    newimage[(y*width+x)*3+0]= (v & 0x0000001f)<<3 | (v & 0x0000001f)>>2;  // blue
                    newimage[(y*width+x)*3+1]= (v & 0x000003e0)>>2 | (v & 0x000003e0)>>7;  // green
                    newimage[(y*width+x)*3+2]= (v & 0x00007c00)>>7 | (v & 0x00007c00)>>12; // red
                    break;
                case 12:            // 16-bit, format 565 (RRRRRGGG GGGBBBBB)
                    v= bits[y*dib_bytesperrow+x*2+0] | (bits[y*dib_bytesperrow+x*2+1]<<8);
                    newimage[(y*width+x)*3+0]= (v & 0x0000001f)<<3 | (v & 0x0000001f)>>2;  // blue
                    newimage[(y*width+x)*3+1]= (v & 0x000007e0)>>3 | (v & 0x000007e0)>>9;  // green
                    newimage[(y*width+x)*3+2]= (v & 0x0000f800)>>8 | (v & 0x0000f800)>>13; // red
                    break;
                case 20:  case 21:  // 32-bit, every 4th byte wasted (b g r x)
                    newimage[(y*width+x)*3+0]= bits[y*dib_bytesperrow+x*4+0]; // blue
                    newimage[(y*width+x)*3+1]= bits[y*dib_bytesperrow+x*4+1]; // green
                    newimage[(y*width+x)*3+2]= bits[y*dib_bytesperrow+x*4+2]; // red
                    break;
                }
            }
        }

        for(i=0;i<height;i++) {
            if(topdown) 
                row_pointers[i]= &newimage[i*width*3];
            else
                row_pointers[height-1-i]= &newimage[i*width*3];
        }
        png_write_image(png_ptr, row_pointers);

        GlobalFree(newimage);
        newimage=NULL;
    }
    else if(iscompressed) {
        newimage= uncompress_dib((LPBITMAPINFO)d2pp->lpdib, headersize+bfsize+palsize, bits);
        if(!newimage) { rv=PNGD_E_NOMEM; goto abort; }
        for(i=0;i<height;i++) {
            if(topdown) 
                row_pointers[i]= &newimage[i*dib_bytesperrow];
            else
                row_pointers[height-1-i]= &newimage[i*dib_bytesperrow];
        }
        png_write_image(png_ptr, row_pointers);

        GlobalFree(newimage);
        newimage=NULL;
    }
    else {
        for(i=0;i<height;i++) {
            if(topdown) 
                row_pointers[i]= &bits[i*dib_bytesperrow];
            else
                row_pointers[height-1-i]= &bits[i*dib_bytesperrow];
        }
        png_write_image(png_ptr, row_pointers);
    }

    GlobalFree((VOID*)row_pointers);
    row_pointers=NULL;

    if(d2pp->software) {
        text_ptr[0].key = "Software";
        text_ptr[0].text = d2pp->software;
        text_ptr[0].compression = PNG_TEXT_COMPRESSION_NONE;
        png_set_text(png_ptr, info_ptr, text_ptr, 1);
    }

    png_write_end(png_ptr, info_ptr);

    rv=PNGD_E_SUCCESS;

abort:
    if(png_ptr) png_destroy_write_struct(&png_ptr, &info_ptr);
    if(fp) fclose(fp);
    if(row_pointers) GlobalFree(row_pointers);
    if(newimage) GlobalFree(newimage);

    return rv;
}

char* pngdib_get_version_string(void)
{
    return PNGDIB_SRC_VERSION_STRING;
}

int pngdib_get_version(void)
{
    return PNGDIB_SRC_VERSION;
}





///////////////////////////////////////////////////////////////


// bmp2png.c
// A sample program that uses the PNGDIB library
// converts a BMP image file to a PNG image file



// This is just a generic function to read a file into a memory block.
int read_bmp_to_mem(char *bmpfn,unsigned char **bmppp, DWORD *fsizep)
{
    HANDLE hfile;
    DWORD fsize;
    unsigned char *fbuf;
    DWORD bytesread;

    hfile=CreateFile(bmpfn,GENERIC_READ,FILE_SHARE_READ,NULL,
        OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
    if(hfile==INVALID_HANDLE_VALUE) return 1;

    fsize=GetFileSize(hfile,NULL);
    if(fsize>0) {
        fbuf=(unsigned char*)GlobalAlloc(GPTR,fsize);
        if(fbuf) {
            if(ReadFile(hfile,(void*)fbuf,fsize,&bytesread,NULL)) {
                if(bytesread==fsize) { 
                    (*bmppp)  = fbuf;
                    (*fsizep) = fsize;
                    CloseHandle(hfile);
                    return 0;   // success
                }
            }
            GlobalFree((void*)fbuf);
        }
    }
    CloseHandle(hfile);
    return 1;  // error
}


/* //////////////////////////////////////////////////////////////////////// //

Actual Beginning of Module code

// //////////////////////////////////////////////////////////////////////// */



extern "C"
{
    int initModuleEx(HWND parentWnd, HINSTANCE dllInst, LPCSTR szPath);
    void quitModule(HINSTANCE dllInst);
}



const char rcsRevision[] = "$Revision: 0.49 $";
const char rcsId[] = "$Id: rymshot.dll v0.49 2002-04-10 09:27 lrddnattor $";


char format_extensions[5][5] =
{
      ".jpg"
    , ".bmp"
    , ".png"
    , ".pnm"
};

bool win_nt;
HWND module_parent;

// general options
int jpeg_quality;
char save_filename[MAX_PATH] = "";
image_format save_format;

// active options
char active_filename[MAX_PATH] = "";
image_format active_format;
int active_quality;

// ftp options
BOOL use_ftp;
char ftp_user[MAX_PATH] = "";
char ftp_password[MAX_PATH] = "";
char ftp_server[MAX_PATH] = "";
char ftp_folder[MAX_PATH] = "";
char ftp_filename[MAX_PATH] = "";
int ftp_quality;
image_format ftp_format;
int ftp_error_treshold;
bool ftp_no_caption;

// thumbnail ftp options
char thumbnail_ftp_folder[MAX_PATH] = "";
char thumbnail_ftp_filename[MAX_PATH] = "";
image_format thumbnail_ftp_format;
int thumbnail_ftp_quality;
bool ftp_no_thumbnail;

// thumbnail options
BOOL use_thumbnail;
char thumbnail_filename[MAX_PATH] = "";
image_format thumbnail_format;
int thumbnail_width;
int thumbnail_height;
int thumbnail_quality;

// caption options
BOOL use_caption;
int caption_x;
int caption_y;
COLORREF caption_color;
char caption_date_format[MAX_PATH] = "";

int screenshot_area(const char * filename
    , int image_width
    , int image_height
    , int quality
    , image_format format
    , bool capture_whole_screen = false
    , bool no_caption = false
    , bool is_thumbnail = false);


int write_PNG_file(const char * filename
    , int image_width = 0
    , int image_height = 0)
{
    char *bmpfn,*pngfn;
    PNGD_D2PINFO d2p;
    int ret;
    unsigned char *bmpp;
    DWORD bmpsize;
    LPBITMAPFILEHEADER lpbmfh;
    char errmsg[100];

    pngfn=const_cast<char *>(filename);

    char tmp[] = "c:/rymshot-4png.bmp";
    if (image_width && image_height)
    {
        screenshot_area(tmp, image_width, image_height, 0, BMP_FORMAT, 1);
    }
    else
    {
        screenshot_area(tmp, 0, 0, 0, BMP_FORMAT);
    }
    bmpfn = tmp;
    if(read_bmp_to_mem(bmpfn,&bmpp, &bmpsize))
    {
        printf("can't read BMP from file\n");
        return 1;
    }
    unlink(tmp);
    ZeroMemory((void*)&d2p,sizeof(PNGD_D2PINFO));
    d2p.structsize=sizeof(PNGD_D2PINFO);

    // The only flag currently defined is PNGD_INTERLACE, which
    // does what you'd expect.
    d2p.flags=0;
    d2p.pngfn=pngfn;

    // The first 14 bytes of a BMP file is a file header, the
    // rest is a DIB.
    d2p.lpdib = (LPBITMAPINFOHEADER)&bmpp[14];

    // This can be left at 0 if you know your DIB is valid.
    d2p.dibsize = bmpsize-14;

    lpbmfh= (LPBITMAPFILEHEADER)bmpp;
    if(lpbmfh->bfOffBits >= bmpsize)
    {
        printf("Corrupt BMP\n");
        return 1;
    }

    // It's usually okay to leave lpbits set to NULL.
    d2p.lpbits = &bmpp[lpbmfh->bfOffBits];

    // This can be left at 0 if you know your DIB is valid.
    d2p.bitssize = bmpsize-lpbmfh->bfOffBits;
    d2p.software = "PNGDIB test app";
    d2p.errmsg = errmsg; strncpy(errmsg,"", 100);

    // returns 0 on success
    if((ret=write_dib_to_png(&d2p)))
    {
        printf("Error: %s (%d)\n", d2p.errmsg, ret);
        return 1;
    }

    GlobalFree(bmpp);

    return 0;
}



void write_JPEG_file (
      const char * filename
    , int quality
    , const JSAMPLE * image_buffer
    , int image_width, int image_height)
{
    struct jpeg_compress_struct cinfo;
    struct jpeg_error_mgr jerr;
    FILE * outfile;
    JSAMPROW row_pointer[1];
    int row_stride;

    cinfo.err = jpeg_std_error(&jerr);
    jpeg_create_compress(&cinfo);

    if (!(outfile = fopen(filename, "wb")))
    {
        MessageBox(0, filename, "can't open file for writing - rymshot", MB_OK);
        jpeg_destroy_compress(&cinfo);
        return;
    }

    jpeg_stdio_dest(&cinfo, outfile);

    cinfo.image_width = image_width;
    cinfo.image_height = image_height;
    cinfo.input_components = 3;
    cinfo.in_color_space = JCS_RGB;
    jpeg_set_defaults(&cinfo);
    jpeg_set_quality(&cinfo, quality, TRUE);
    jpeg_start_compress(&cinfo, TRUE);
    row_stride = image_width * 3;

    while (cinfo.next_scanline < cinfo.image_height)
    {
        int r;
        row_pointer[0] = const_cast<JSAMPLE *>(&image_buffer[cinfo.next_scanline * row_stride]);

        // bgr -> rgb conversion
        for (r = 0; r < row_stride; r+=3)
        {
            const JSAMPLE tmp = row_pointer[0][r];
            row_pointer[0][r] = row_pointer[0][r+2];
            row_pointer[0][r+2] = tmp;
        }
        jpeg_write_scanlines(&cinfo, row_pointer, 1);
    }

    jpeg_finish_compress(&cinfo);
    fclose(outfile);

    jpeg_destroy_compress(&cinfo);
}


void write_BMP_file(const char * filename
    , BITMAPFILEHEADER * pbmFileHeader
    , BITMAPINFO * pbmInfo
    , void * lpvpxldata
    , int image_data_size)
{
    unlink(filename);
    int bmfile = open(filename, O_WRONLY | O_CREAT | O_TRUNC | O_BINARY | O_EXCL, S_IWRITE);
    write(bmfile, pbmFileHeader, sizeof(BITMAPFILEHEADER));
    write(bmfile, pbmInfo, sizeof(BITMAPINFO));
    write(bmfile, lpvpxldata, image_data_size);
    close(bmfile);
}

int screenshot_area(const char * filename
    , int image_width
    , int image_height
    , int quality
    , image_format format
    , bool capture_whole_screen
    , bool no_caption
    , bool is_thumbnail)
{
    void * lpvpxldata;

    BITMAPINFO bmInfo;
    BITMAPFILEHEADER bmFileHeader;

    int image_data_size;
    int window_width;
    int window_height;
    HBITMAP hmemBM;
    HGDIOBJ hOldObj;
    HDC hmemDC;
    int window_bitspixel;
    const HWND foregroundWnd = capture_whole_screen ? 0 : GetForegroundWindow();
    const HDC hScreenDC = GetWindowDC(foregroundWnd);

    if ((capture_whole_screen |= !foregroundWnd))
    {
        window_width = GetDeviceCaps(hScreenDC, HORZRES);
        window_height = GetDeviceCaps(hScreenDC, VERTRES);
    }
    else
    {
        RECT window_rect;
        GetWindowRect(foregroundWnd, &window_rect);
        window_width = window_rect.right - window_rect.left + 1;
        window_height = window_rect.bottom - window_rect.top + 1;
    }

    hmemDC = CreateCompatibleDC(hScreenDC);
    hmemBM = CreateCompatibleBitmap(hScreenDC
        , image_width  ? image_width  : (image_width  = window_width)
        , image_height ? image_height : (image_height = window_height));
    hOldObj = SelectObject(hmemDC, hmemBM);
    SetStretchBltMode(hmemDC, COLORONCOLOR);

    StretchBlt(hmemDC, 0, 0, image_width, image_height
        , hScreenDC, 0, 0, window_width, window_height
        , SRCCOPY | (win_nt && CAPTUREBLT));

    if (use_caption && !no_caption)
    {
        time_t now;
        size_t len;
        char str[MAX_PATH];
        SetBkMode(hmemDC, TRANSPARENT);
        SetTextColor(hmemDC, caption_color);
        now = time(NULL);
        len = strftime(str, MAX_PATH, caption_date_format, localtime(&now));
        TextOut(hmemDC, caption_x, caption_y, str, len);
    }

    SelectObject(hmemDC, hOldObj);

    if (format == JPEG_FORMAT)
    {
        window_bitspixel = 24;
    }
    else
    {
        window_bitspixel = GetDeviceCaps(hScreenDC, BITSPIXEL);
    };
    ReleaseDC(0, hScreenDC);

    switch (window_bitspixel)
    {
        case 16:
        case 24:
        case 32:
            image_data_size = image_width * image_height * (window_bitspixel / 8);
        break;

        case 1:
        case 4:
        case 8:
        default:
            MessageBox(0, "unsupported bitsperpixel", "rymshot", MB_OK);
            DeleteObject(hmemBM);
            DeleteDC(hmemDC);
            return 1;
    }

    // fill .bmp - structures
    bmInfo.bmiHeader.biSize = sizeof(bmInfo.bmiHeader);
    bmInfo.bmiHeader.biWidth = image_width;

    switch (format)
    {
    case JPEG_FORMAT:
        bmInfo.bmiHeader.biHeight = -image_height;
    break;

    case PNM_FORMAT:
    case PNG_FORMAT:
        // do nothing
    break;
    
    case BMP_FORMAT:
        bmInfo.bmiHeader.biHeight = image_height;
    break;
    }
    
    bmInfo.bmiHeader.biPlanes = 1;
    bmInfo.bmiHeader.biBitCount = window_bitspixel;
    bmInfo.bmiHeader.biCompression = BI_RGB;    
    bmInfo.bmiHeader.biSizeImage = 0;
    bmInfo.bmiHeader.biXPelsPerMeter = 0;
    bmInfo.bmiHeader.biYPelsPerMeter = 0;
    bmInfo.bmiHeader.biClrUsed = 0;
    bmInfo.bmiHeader.biClrImportant = 0;

    bmFileHeader.bfType = 19778;
    bmFileHeader.bfSize = (image_data_size) + sizeof(BITMAPINFO) + sizeof(BITMAPFILEHEADER);
    bmFileHeader.bfReserved1 = 0;
    bmFileHeader.bfReserved2 = 0;
    bmFileHeader.bfOffBits = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFO);

    lpvpxldata = malloc(image_data_size);

    // copy bitmap data into global memory
    GetDIBits(hmemDC, hmemBM, 0, image_height, lpvpxldata, &bmInfo, DIB_RGB_COLORS);

    switch (format)
    {
    case JPEG_FORMAT:
        write_JPEG_file(filename, quality, (const JSAMPLE *)lpvpxldata, image_width, image_height);
    break;
    
    case BMP_FORMAT:
        write_BMP_file(filename, &bmFileHeader, &bmInfo, (JSAMPLE *)lpvpxldata, image_data_size);
    break;
    
    case PNG_FORMAT:
        if (capture_whole_screen)
        {
            write_PNG_file(filename, image_width, image_height);
        }
        else
        {
            write_PNG_file(filename);
        }
    break;

    case PNM_FORMAT:
        // do nothing
    break;
    }

    DeleteObject(hmemBM);
    DeleteDC(hmemDC);
    free(lpvpxldata);

    return 0;
}

int screenshot_desktop(const char * filename
    , int image_width
    , int image_height
    , int quality
    , image_format format
    , bool no_caption
    , bool is_thumbnail)
{
    return screenshot_area(filename, image_width, image_height, quality, format, 1, no_caption, is_thumbnail);
}

void bangRymshotactive(HWND hCaller, LPCSTR szArgs)
{
    screenshot_area(active_filename, 0, 0, active_quality, active_format);
}


void bangRymshotnotn(HWND hCaller, LPCSTR szArgs)
{
    screenshot_desktop(save_filename, 0, 0, jpeg_quality, save_format);
}

void bangRymshottn(HWND hCaller, LPCSTR szArgs)
{
    screenshot_desktop(thumbnail_filename, thumbnail_width, thumbnail_height, thumbnail_quality, thumbnail_format);
}

void bangRymshot(HWND hCaller, LPCSTR szArgs)
{
    if (use_thumbnail)
    {
        screenshot_desktop(thumbnail_filename, thumbnail_width, thumbnail_height, thumbnail_quality, thumbnail_format);
    }
    screenshot_desktop(save_filename, 0, 0, jpeg_quality, save_format);
}


bool is_valid_ftp_password(const char * pw)
{
    return (pw && strncmp(pw, "", MAX_PATH));
}

int send_to_ftp(const char * source_file, bool thumbnail = false)
{
    HINTERNET hNet_0;
    HINTERNET hNet_1;
    BOOL res;
    char out_url[MAX_PATH];

    hNet_0 = InternetOpen(NULL, INTERNET_OPEN_TYPE_DIRECT, NULL, NULL, 0);
    if (!hNet_0)
    {
        return 1;
    }

    hNet_1 = InternetConnect(hNet_0
        , ftp_server
        , INTERNET_DEFAULT_FTP_PORT
        , ftp_user
        , ftp_password
        , INTERNET_SERVICE_FTP, 0, 0);
    if (!hNet_1)
    {
        InternetCloseHandle(hNet_0);
        return 2;
    }

    if (thumbnail)
    {
        strncpy(out_url, thumbnail_ftp_folder, MAX_PATH);
        strncat(out_url, thumbnail_ftp_filename, MAX_PATH);//young
    }
    else
    {
        strncpy(out_url, ftp_folder, MAX_PATH);
        strncat(out_url, ftp_filename, MAX_PATH);
    }

    res = FtpPutFile(hNet_1
        , source_file
        , out_url
        , FTP_TRANSFER_TYPE_BINARY
        , 0);

    InternetCloseHandle(hNet_1);
    InternetCloseHandle(hNet_0);

    return res ? 0 : 3;
}


void bangRymshotftppassword(HWND hCaller, LPCSTR szArgs)
{
    char arg[MAX_PATH];
    char* tokens[1] = {arg};

//  if(LCTokenize(szArgs, tokens, 1, NULL) == 1)
    if(LCTokenize(szArgs, tokens, 1, NULL))
        strncpy(ftp_password, arg, MAX_PATH);
}



void bangRymshotftp(HWND hCaller, LPCSTR szArgs)
{
    static int consecutive_errors = 0;
    char tmp[] = "c:/rymshot.tmp";
    char tmp_tn[] = "c:/rymshot-tn.tmp";

    if (consecutive_errors > ftp_error_treshold) return;

    if (!is_valid_ftp_password(ftp_password))
    {
        ++consecutive_errors;

        MessageBox(0
            , "use !rymshotftppassword password to set your ftp password"
            , "rymshot"
            , MB_OK);
        return;
    }

    screenshot_desktop(tmp, 0, 0, ftp_quality, ftp_format, ftp_no_caption);
    consecutive_errors = send_to_ftp(tmp) ? consecutive_errors + 1 : 0;
    unlink(tmp);

    if (!ftp_no_thumbnail)
    {
        screenshot_desktop(tmp_tn
            , thumbnail_width
            , thumbnail_height
            , thumbnail_ftp_quality
            , thumbnail_ftp_format
            , ftp_no_caption);

        consecutive_errors = send_to_ftp(tmp_tn, true) ? consecutive_errors + 1 : 0;
        unlink(tmp_tn);
    }
}

void bangRymshotabout(HWND hCaller, LPCSTR szArgs)
{
	const size_t max_len = 256;
	char str[max_len];
	char buf[max_len];

	strncpy(str, rcsId, max_len);

	snprintf(buf
		, max_len
		, "\nfilename: %s\tquality: %d\tformat: %s"
		, save_filename
		, jpeg_quality
		, format_extensions[save_format]);

	if ((strlen(str) + strlen(buf)) < max_len)
	{
		strncat(str, buf, max_len);
	}
	else
	{
		MessageBox(0, str, "rymshot.dll", MB_OK|MB_ICONINFORMATION);
		strncpy(str, buf, max_len);
	}


	snprintf(buf
		, max_len
		, "\nactive: %s\tquality: %d\tformat: %s"
		, active_filename
		, active_quality
		, format_extensions[active_format]);
	if ((strlen(str) + strlen(buf)) < max_len)
	{
		strncat(str, buf, max_len);
	}
	else
	{
		MessageBox(0, str, "rymshot.dll", MB_OK|MB_ICONINFORMATION);
		strncpy(str, buf, max_len);
	}

	if (use_thumbnail)
	{
		snprintf(buf
			, max_len
			, "\nthumbnail: %s\tquality: %d\tformat: %s\n\twidth: %d\theight: %d"
			, thumbnail_filename
			, thumbnail_quality
			, format_extensions[thumbnail_format]
			, thumbnail_width
			, thumbnail_height
			);
		if ((strlen(str) + strlen(buf)) < max_len)
		{
			strncat(str, buf, max_len);
		}
		else
		{
			MessageBox(0, str, "rymshot.dll", MB_OK|MB_ICONINFORMATION);
			strncpy(str, buf, max_len);
		}
	}

	if (use_ftp)
	{
		snprintf(buf
			, max_len
			, "\nftp: %s\tquality: %d\tformat: %s\tserver: %s \tusername: %s"
			, ftp_filename
			, ftp_quality
			, format_extensions[ftp_format]
			, ftp_server
			, ftp_user);
		if ((strlen(str) + strlen(buf)) < max_len)
		{
			strncat(str, buf, max_len);
		}
		else
		{
			MessageBox(0, str, "rymshot.dll", MB_OK|MB_ICONINFORMATION);
			strncpy(str, buf, max_len);
		}
	}

	if (use_thumbnail && use_ftp && !ftp_no_thumbnail)
	{
		snprintf(buf
			, max_len
			, "\nftp thumbnail: %s\tquality: %d\tformat: %s"
			, thumbnail_ftp_filename
			, thumbnail_ftp_quality
			, format_extensions[thumbnail_ftp_format]);
		if ((strlen(str) + strlen(buf)) < max_len)
		{
			strncat(str, buf, max_len);
		}
		else
		{
			MessageBox(0, str, "rymshot.dll", MB_OK|MB_ICONINFORMATION);
			strncpy(str, buf, max_len);
		}
	}

	if (use_caption)
	{
		snprintf(buf
			, max_len
			, "\ncaption: %s\tx: %d\ty: %d"
			, caption_date_format
			, caption_x
			, caption_y);
		if ((strlen(str) + strlen(buf)) < max_len)
		{
			strncat(str, buf, max_len);
		}
		else
		{
			MessageBox(0, str, "rymshot.dll", MB_OK|MB_ICONINFORMATION);
			strncpy(str, buf, max_len);
		}
	}
	MessageBox(module_parent, str, "rymshot.dll", MB_OK|MB_ICONINFORMATION);
//	MessageBox(0, str, "rymshot.dll", MB_OK|MB_ICONINFORMATION);

//	snprintf(buf, max_len, "%d", strlen(str));
//	MessageBox(0, buf, "rymshot.dll", MB_OK|MB_ICONINFORMATION);
}

image_format get_format(const char * str)
{
    if (!strncmp(str, "bmp", MAX_PATH))
        return BMP_FORMAT;

    if (!strncmp(str, "png", MAX_PATH))
        return PNG_FORMAT;

//  if (!strncmp(str, "pnm", MAX_PATH))
//      return PNM_FORMAT;

    return JPEG_FORMAT;
}


void read_config(void)
{
    char save_format_l[MAX_PATH];
    char active_format_l[MAX_PATH];
    char buf[MAX_PATH];
    char ftp_format_l[MAX_PATH];
    bool local4ftp;

    if ((use_caption = GetRCBoolDef("rymshotcaption", FALSE)))
    {
        caption_x = GetRCInt("rymshotcaptionx", 5);
        caption_y = GetRCInt("rymshotcaptiony", 15);
        caption_color = GetRCColor("rymshotcaptioncolor", 0);
        GetRCString("rymshotcaptiondateformat"
            , caption_date_format
            , "by rymshot: %c"
            , MAX_PATH);
    }

    GetRCString("rymshotfilename", save_filename, "rymshot", MAX_PATH);
    jpeg_quality = GetRCInt("rymshotquality", 50);

    GetRCString("rymshotformat", save_format_l, "jpg", MAX_PATH);

    save_format = get_format(save_format_l);
    if (save_format == JPEG_FORMAT)
    {
        strncpy(save_format_l, "jpg", MAX_PATH);
    }

    if ((use_ftp = GetRCBoolDef("rymshotftp", FALSE)))
    {
        local4ftp = GetRCBoolDef("rymshotftpuselocal", FALSE);

        if (use_caption)
        {
            ftp_no_caption = GetRCBoolDef("rymshotftpnocaption", FALSE);
        }

        GetRCString("rymshotftpformat", ftp_format_l, save_format_l, MAX_PATH);
        ftp_format = get_format(ftp_format_l);
        if (ftp_format == JPEG_FORMAT)
        {
            strncpy(ftp_format_l, "jpg", MAX_PATH);
        }

        ftp_error_treshold = GetRCInt("rymshotftperrortreshold", 20);

        ftp_quality = GetRCInt("rymshotftpquality", jpeg_quality);

        GetRCString("rymshotftpuser", ftp_user, NULL, MAX_PATH);
        GetRCString("rymshotftppassword", ftp_password, NULL, MAX_PATH);
        GetRCString("rymshotftpserver", ftp_server, NULL, MAX_PATH);
        GetRCString("rymshotftpfolder", ftp_folder, NULL, MAX_PATH);

        if (local4ftp)
        {
            strncat(ftp_filename, save_filename, MAX_PATH);
        }
        else
        {
            GetRCString("rymshotftpfile", ftp_filename, NULL, MAX_PATH);
        }
        use_ftp = !(!ftp_user[0] | !ftp_server[0] | !ftp_filename[0]);
    }

    if ((use_thumbnail = GetRCBoolDef("rymshotthumbnail", FALSE)))
    {
        char thumbnail_format_l[MAX_PATH];
        char buf[MAX_PATH];
        strncpy(buf, save_filename, MAX_PATH);
        strncat(buf, "-tn", MAX_PATH);
        GetRCString("rymshottnfilename", thumbnail_filename, buf, MAX_PATH);
        thumbnail_width = GetRCInt("rymshottnwidth", 200);
        thumbnail_height = GetRCInt("rymshottnheight", 150);
        thumbnail_quality = GetRCInt("rymshottnquality", jpeg_quality);
        GetRCString("rymshottnformat", thumbnail_format_l, save_format_l, MAX_PATH);
        thumbnail_format = get_format(thumbnail_format_l);
        strncat(thumbnail_filename, format_extensions[thumbnail_format], MAX_PATH);

        if (use_ftp)
        {
            char thumbnail_ftp_format_l[MAX_PATH];
            strncpy(buf, ftp_filename, MAX_PATH);
            strncat(buf, "-tn", MAX_PATH);
            GetRCString("rymshottnftpfile", thumbnail_ftp_filename, buf, MAX_PATH);
            GetRCString("rymshottnftpfolder", thumbnail_ftp_folder, ftp_folder, MAX_PATH);
            thumbnail_ftp_quality = GetRCInt("rymshottnftpquality", ftp_quality);
            ftp_no_thumbnail = GetRCBoolDef("rymshotftpnothumbnail", FALSE);
            GetRCString("rymshottnftpformat", thumbnail_ftp_format_l, ftp_format_l, MAX_PATH);
            thumbnail_ftp_format = get_format(thumbnail_ftp_format_l);
            use_ftp = !!thumbnail_ftp_filename[0];
            strncat(thumbnail_ftp_filename, format_extensions[thumbnail_ftp_format], MAX_PATH);
        }
    }

    if (use_ftp)
    {
        strncat(ftp_filename, format_extensions[ftp_format], MAX_PATH);
    }

    strncpy(buf, save_filename, MAX_PATH);
    strncat(buf, "-active", MAX_PATH);
    GetRCString("rymshotactivefilename", active_filename, buf, MAX_PATH);
    active_quality = GetRCInt("rymshotactivequality", jpeg_quality);
    GetRCString("rymshotactiveformat", active_format_l, save_format_l, MAX_PATH);
    active_format = get_format(active_format_l);

    strncat(active_filename, format_extensions[active_format], MAX_PATH);
    strncat(save_filename, format_extensions[save_format], MAX_PATH);
}

int initModuleEx(HWND parentWnd, HINSTANCE dllInst, LPCSTR szPath)
{
	module_parent = parentWnd;
    win_nt = !HIBYTE(LOWORD(GetVersion()));
    read_config();

    AddBangCommand("!rymshot", bangRymshot);
    AddBangCommand("!rymshotactive", bangRymshotactive);
    AddBangCommand("!rymshotabout", bangRymshotabout);

    if (use_thumbnail)
    {
        AddBangCommand("!rymshottn", bangRymshottn);
        AddBangCommand("!rymshotnotn", bangRymshotnotn);
    }

    if (use_ftp)
    {
        AddBangCommand("!rymshotftp", bangRymshotftp);
        AddBangCommand("!rymshotftppassword", bangRymshotftppassword);
    }

    return 0;
}

void quitModule(HINSTANCE dllInst)
{
    RemoveBangCommand("!rymshot");
    RemoveBangCommand("!rymshotabout");
    RemoveBangCommand("!rymshotactive");
    if (use_thumbnail)
    {
        RemoveBangCommand("!rymshotnotn");
        RemoveBangCommand("!rymshottn");
    }

    if (use_ftp)
    {
        RemoveBangCommand("!rymshotftp");
        RemoveBangCommand("!rymshotftppassword");
    }
}


